@@ -6,8 +6,7 @@ import math
 from fontTools.misc.transform import Transform, Identity
 from fontTools.misc.arrayTools import unionRect
 from fontTools.ttLib import TTFont
-from fontTools.ttLib.tables.otTables import CompositeMode, PaintFormat, VariableValue
-from fontTools.ttLib.tables.otConverters import VarF2Dot14, VarFixed
+from fontTools.ttLib.tables.otTables import CompositeMode, PaintFormat
 from fontTools.varLib.varStore import VarStoreInstancer
 import uharfbuzz as hb
 
@@ -46,9 +45,13 @@ class BlackRendererFont:
                 colrTable = colrTable.table
                 self.colrV1Glyphs = {
                     glyph.BaseGlyph: glyph
-                    for glyph in colrTable.BaseGlyphV1List.BaseGlyphV1Record
+                    for glyph in colrTable.BaseGlyphList.BaseGlyphPaintRecord
                 }
-                self.colrLayersV1 = colrTable.LayerV1List
+                if colrTable.ClipList is None:
+                    self.clipBoxes = None
+                else:
+                    self.clipBoxes = colrTable.ClipList.clips
+                self.colrLayersV1 = colrTable.LayerList
                 if colrTable.VarStore is not None:
                     self.instancer = VarStoreInstancer(
                         colrTable.VarStore, self.ttFont["fvar"].axes
@@ -98,9 +101,13 @@ class BlackRendererFont:
         return self.colrV1Glyphs.keys()
 
     def getGlyphBounds(self, glyphName):
-        if glyphName in self.colrV1Glyphs or glyphName not in self.colrV0Glyphs:
+        if glyphName in self.colrV1Glyphs:
             bounds = self._getGlyphBounds(glyphName)
-        else:
+            if self.clipBoxes is not None:
+                box = self.clipBoxes.get(glyphName)
+                if box is not None:
+                    bounds = box.xMin, box.yMin, box.xMax, box.yMax
+        elif glyphName in self.colrV0Glyphs:
             # For COLRv0, we take the union of all layer bounds
             bounds = None
             for layer in self.colrV0Glyphs[glyphName]:
@@ -109,6 +116,8 @@ class BlackRendererFont:
                     bounds = layerBounds
                 else:
                     bounds = unionRect(layerBounds, bounds)
+        else:
+            bounds = self._getGlyphBounds(glyphName)
         return bounds
 
     def drawGlyph(self, glyphName, canvas, *, palette=None, textColor=(0, 0, 0, 1)):
@@ -116,6 +125,7 @@ class BlackRendererFont:
             palette = self.palettes[0]
         self.currentPalette = palette
         self.textColor = textColor
+        self._recursionCheck = set()
 
         glyph = self.colrV1Glyphs.get(glyphName)
         if glyph is not None:
@@ -142,7 +152,13 @@ class BlackRendererFont:
             canvas.drawPathSolid(path, self._getColor(layer.colorID, 1))
 
     def _drawGlyphCOLRv1(self, glyph, canvas):
-        self._drawPaint(glyph.Paint, canvas)
+        if glyph.BaseGlyph in self._recursionCheck:
+            raise RecursionError(f"Glyph '{glyph.BaseGlyph}' references itself")
+        self._recursionCheck.add(glyph.BaseGlyph)
+        try:
+            self._drawPaint(glyph.Paint, canvas)
+        finally:
+            self._recursionCheck.remove(glyph.BaseGlyph)
 
     # COLRv1 Paint dispatch
 
@@ -171,7 +187,7 @@ class BlackRendererFont:
                     self._drawPaint(self.colrLayersV1.Paint[i], canvas)
 
     def _drawPaintSolid(self, paint, canvas):
-        color = self._getColor(paint.Color.PaletteIndex, paint.Color.Alpha)
+        color = self._getColor(paint.PaletteIndex, paint.Alpha)
         canvas.drawPathSolid(self.currentPath, color)
 
     def _drawPaintLinearGradient(self, paint, canvas):
@@ -247,6 +263,11 @@ class BlackRendererFont:
         self._applyTransform(transform, paint.Paint, canvas)
 
     def _drawPaintRotate(self, paint, canvas):
+        transform = Transform()
+        transform = transform.rotate(math.radians(paint.angle))
+        self._applyTransform(transform, paint.Paint, canvas)
+
+    def _drawPaintRotateAroundCenter(self, paint, canvas):
         transform = Transform()
         transform = transform.translate(paint.centerX, paint.centerY)
         transform = transform.rotate(math.radians(paint.angle))
@@ -254,6 +275,13 @@ class BlackRendererFont:
         self._applyTransform(transform, paint.Paint, canvas)
 
     def _drawPaintSkew(self, paint, canvas):
+        transform = Transform()
+        transform = transform.skew(
+            math.radians(paint.xSkewAngle), math.radians(paint.ySkewAngle)
+        )
+        self._applyTransform(transform, paint.Paint, canvas)
+
+    def _drawPaintSkewAroundCenter(self, paint, canvas):
         transform = Transform()
         transform = transform.translate(paint.centerX, paint.centerY)
         transform = transform.skew(
@@ -263,10 +291,26 @@ class BlackRendererFont:
         self._applyTransform(transform, paint.Paint, canvas)
 
     def _drawPaintScale(self, paint, canvas):
-        # https://github.com/googlefonts/colr-gradients-spec/issues/279
+        transform = Transform()
+        transform = transform.scale(paint.scaleX, paint.scaleY)
+        self._applyTransform(transform, paint.Paint, canvas)
+
+    def _drawPaintScaleAroundCenter(self, paint, canvas):
         transform = Transform()
         transform = transform.translate(paint.centerX, paint.centerY)
-        transform = transform.scale(paint.xScale, paint.yScale)
+        transform = transform.scale(paint.scaleX, paint.scaleY)
+        transform = transform.translate(-paint.centerX, -paint.centerY)
+        self._applyTransform(transform, paint.Paint, canvas)
+
+    def _drawPaintScaleUniform(self, paint, canvas):
+        transform = Transform()
+        transform = transform.scale(paint.scale, paint.scale)
+        self._applyTransform(transform, paint.Paint, canvas)
+
+    def _drawPaintScaleUniformAroundCenter(self, paint, canvas):
+        transform = Transform()
+        transform = transform.translate(paint.centerX, paint.centerY)
+        transform = transform.scale(paint.scale, paint.scale)
         transform = transform.translate(-paint.centerX, -paint.centerY)
         self._applyTransform(transform, paint.Paint, canvas)
 
@@ -373,7 +417,7 @@ class BlackRendererFont:
     def _readColorLine(self, colorLineTable):
         return _normalizeColorLine(
             [
-                (cs.StopOffset, self._getColor(cs.Color.PaletteIndex, cs.Color.Alpha))
+                (cs.StopOffset, self._getColor(cs.PaletteIndex, cs.Alpha))
                 for cs in colorLineTable.ColorStop
             ]
         )
@@ -408,6 +452,13 @@ def _normalizeColorLine(colorLine):
         # interpolation. By setting minStop and maxStop to 0 and 1,
         # we at least won't mess up the parameters.
         # FIXME: should the gradient simply not be drawn in this case?
+        # 1. If there's only one color stop, we can just draw a solid
+        # color.
+        # 2. If there are > 1 color stops all at the same color stop
+        # offset, things get interesting: everything "to the left"
+        # should get color[0], everything "to the right" should get
+        # color[-1]. If there are more than 2 color stops, all but the
+        # first and last should be ignored.
         minStop, maxStop = 0, 1
     return minStop, maxStop, colorLine
 
@@ -435,10 +486,10 @@ def axisValuesToLocation(normalizedAxisValues, axisTags):
     }
 
 
-_conversionFactors = {
-    VarF2Dot14: 1 / (1 << 14),
-    VarFixed: 1 / (1 << 16),
-}
+# _conversionFactors = {
+#     VarF2Dot14: 1 / (1 << 14),
+#     VarFixed: 1 / (1 << 16),
+# }
 
 
 class PaintVarWrapper:
@@ -452,16 +503,17 @@ class PaintVarWrapper:
 
     def __getattr__(self, attrName):
         value = getattr(self._wrappedPaint, attrName)
-        if isinstance(value, VariableValue):
-            if value.varIdx != 0xFFFFFFFF:
-                factor = _conversionFactors.get(
-                    type(self._wrappedPaint.getConverterByName(attrName)), 1
-                )
-                value = value.value + self._instancer[value.varIdx] * factor
-            else:
-                value = value.value
-        elif type(value).__name__.startswith("Var"):
-            value = PaintVarWrapper(value, self._instancer)
-        elif isinstance(value, (list, UserList)):
-            value = [PaintVarWrapper(item, self._instancer) for item in value]
+        raise NotImplementedError("This code is currently not working")
+        # if isinstance(value, VariableValue):
+        #     if value.varIdx != 0xFFFFFFFF:
+        #         factor = _conversionFactors.get(
+        #             type(self._wrappedPaint.getConverterByName(attrName)), 1
+        #         )
+        #         value = value.value + self._instancer[value.varIdx] * factor
+        #     else:
+        #         value = value.value
+        # elif type(value).__name__.startswith("Var"):
+        #     value = PaintVarWrapper(value, self._instancer)
+        # elif isinstance(value, (list, UserList)):
+        #     value = [PaintVarWrapper(item, self._instancer) for item in value]
         return value
